#include "analysis.h"
#include "unitConversion.h"
#include <math.h>

//---------------------------------------------------------------------
// Helper functions and constants

#ifndef	M_PI
const double M_PI = 2*acos(0.0);
#endif

static inline	double	sqr(double a) { return (a)*(a); }
static inline double dist2(const CS_REAL3 &a, const CS_REAL3 &b)
{
    double dx = a.s[0] - b.s[0];
    double dy = a.s[1] - b.s[1];
    double dz = a.s[2] - b.s[2];
    return dx*dx + dy*dy + dz*dz;
}

static inline double dist(const CS_REAL3 &a, const CS_REAL3 &b)
{
    return sqrt( dist2(a,b) );
}

static CS_REAL3 find_mean(const std::vector<CS_REAL3> &pts)
{
    // If we have no points, return zero.
    CS_REAL3 mean;
    mean.s[0] = (0);
    mean.s[1] = (0);
    mean.s[2] = (0);
    if (pts.size() == 0) {
	return mean;
    }

    // Compute the mean location of the all the points
    size_t i;
    for (i = 0; i < pts.size(); i++) {
	mean.s[0] += pts[i].s[0];
	mean.s[1] += pts[i].s[1];
	mean.s[2] += pts[i].s[2];
    }
    mean.s[0] /= pts.size();
    mean.s[1] /= pts.size();
    mean.s[2] /= pts.size();

    return mean;
}


//---------------------------------------------------------------------
// Functions declared in the header

double compute_MSD(const std::vector<CS_REAL3> &points, unsigned stride)
{
    double sum = 0;
    int count = static_cast<int>(points.size() - stride);
    for (int i = 0; i < count ; i++) {
        for (int j = 0; j < 3; j++) {
            sum += sqr(points[i].s[j] - points[i+stride].s[j]);
        }
    }
    if (count > 0) {
        sum /= count;
    }
    //printf("XXX MSD based on %d estimates (%d points)\n", count, static_cast<int>(points.size()));
    return sum;
}

double expected_MSD(
        double temperature_Kelvin           // Of the medium
        , double tau_seconds                // Time interval between position measures
        , double viscosity_Pascal_seconds   // Of the medium
        , double radius_meters              // Of a spherical particle that is diffusing
        )
{
    return tau_seconds * (Boltzmann_constant * temperature_Kelvin )
            / (M_PI * viscosity_Pascal_seconds * radius_meters);
}


double compute_RoG(const std::vector<CS_REAL3> &points)
{
    // If we don't have any points, we return a negative radius,
    // which indicates an error.
    if (points.size() == 0) { return -1; }

    // Compute the mean location of the points.
    CS_REAL3 mean = find_mean(points);

    // Compute the average squared distance from the mean
    size_t i;
    double sum = 0.0;
    for (i = 0; i < points.size(); i++) {
	sum += dist2(mean, points[i]);
    }
    return sqrt( sum / points.size() );
}

double expected_persistence_length(
    double temperature_Kelvin           // Of the medium
    , double Youngs_modulus_Pascal      // Of the DNA segment
    , double Radius_meters              // Of the DNA segment
    )
{
    // Compute bending rigidity E * I
    double EI = dna_bending_rigidity(Youngs_modulus_Pascal, Radius_meters);
    return EI / ( Boltzmann_constant * temperature_Kelvin);
}

double expected_RoG(
    double Lc_meters        // Contour length of the simulated segment of DNA
    , double Lp_meters      // Persistence length of the DNA segment
    )
{
    return sqrt(Lp_meters * Lc_meters / 6);
}

extern double compute_squared_end_to_end(const std::vector<CS_REAL3> &points)
{
    // If we don't have at least two points, return -1, which
    // indicates an error.
    if (points.size() < 2) { return -1; }

    return dist2(points[0], points[points.size()-1]);
}

double compute_contour_length(const std::vector<CS_REAL3> &points)
{
    // If we don't have at least two points, return -1, which
    // indicates an error.
    if (points.size() < 2) { return -1; }

    double length = 0;
    size_t i;
    for (i = 1; i < points.size(); i++) {
	length += dist(points[i-1], points[i]);
    }

    return length;
}

double Rog_collision_multiplier(unsigned number_of_segments)
{
    // For Ideal chain Rg= b*N^1/2   
    //  (where b is the length of a segment and N is the number of segments)
    // For real chain Rg = b*N^3/5
    // So the multiplier  Rg(real)/Rg(ideal)= N^(1/10)

    return pow(static_cast<double>(number_of_segments), 1.0/10.0);
}
