#pragma once

#include "chromoTypes.h"
#include <vector>

//---------------------------------------------------------------------
// Compute the mean of the squared displacements of the list of points.
// This is done with the specified stride.
extern double compute_MSD(const std::vector<CS_REAL3> &points, unsigned stride = 1);

//---------------------------------------------------------------------
// Compute the expected MSD for a moving particle depending on its
// parameters and the medium it is moving in.
extern double expected_MSD(
                double temperature_Kelvin           // Of the medium
                , double tau_seconds                // Time interval between position measures
                , double viscosity_Pascal_seconds   // Of the medium
                , double radius_meters              // Of a spherical particle that is diffusing
                );

//---------------------------------------------------------------------
// Compute the radius of gyration of the set of points.  This is the root mean
// of the squared distances from the center of mass to each point in the
// chain.
extern double compute_RoG(const std::vector<CS_REAL3> &points);

//---------------------------------------------------------------------
// Compute the squared end-to-end distance of the set of points.
extern double compute_squared_end_to_end(const std::vector<CS_REAL3> &points);

//---------------------------------------------------------------------
// Compute the contour length of the set of points.
extern double compute_contour_length(const std::vector<CS_REAL3> &points);

//---------------------------------------------------------------------
// Compute the expected radius of gyration given the characteristics
// of the simulation.  This is for a single chain of beads with springs
// and hinges.
/*
	The expected end-to-end distance is sqrt(Lp*Lc), where Lc
	is the contour length (the distance traced along the chain
	from one end to the other) and Lp is the persistence length
	(about 50nm for B-form DNA).

        Rg = sqrt(<R(from center of mass)^2>)
        Lp = persistence length.  Lc = contour length.
        Lp*Lc = <R(end to end)^2>,  Also, Lp = E*I / (kB*T),
        which turns out to be when the correlation of the cosine of the angles goes below 1/e.
        It turns out that Rg^2 = Lp*Lc / 6.
	The expected radius of gyration is sqrt(Lp*Lc/6).
*/
extern double expected_persistence_length(
                double temperature_Kelvin           // Of the medium
                , double Youngs_modulus_Pascal      // Of the DNA segment
                , double Radius_meters              // Of the DNA segment
                );
extern double expected_RoG(
                double Lc_meters        // Contour length of the simulated segment of DNA
                , double Lp_meters      // Persistence length of the DNA segment
                );
// Scales the above expected_RoG when there are collisions, based on the
// number of segments in the chain.
extern double Rog_collision_multiplier(unsigned number_of_segments);
