// chromoSim.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "chromoSim.h"
#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>


// simulate time step forward
bool chromoSim::simulateTimeStep( )
{
    if (!doing_okay()) { return false; }

    // integrate forward a step
    solver->integrateStep( );

    // make sure the integration step completes
    if (CheckErr( clFlush( queue ), "clFlush (after integration)" )) {
	broken = true;
	return false;
    }
    if (CheckErr( clFinish( queue ), "clFinish (after integration)" )) {
	broken = true;
	return false;
    }

    // update the fact that we've run a time step
    ++timeStepsMade;
    return true;
}

// profile randomness
bool chromoSim::profileRandom( int numSteps, bool do_integers, std::string fileName )
{
    if (!doing_okay()) { return false; }

    // save the random numbers generated in the first numSteps steps
    // to the path specified

    forceRandomGPU * randomGenerator = dynamic_cast<forceRandomGPU *>( gBrownian );
    bool doIntegers = ( randomGenerator != NULL ) && do_integers;

    if ( doIntegers ) {
        // allow profiling of integers
        randomGenerator->initializeProfileIntegers( );

        std::cout << "Saving random integers" << std::endl;
    } else {
        std::cout << "Saving normally distributed random variables" << std::endl;
    }
    // loop over all the steps
    for (int i = 0; i < numSteps; ++i) {
        if (i % 1000 == 0) {
            // update that we are still working
            std::cout << "Simulated " << i << " steps out of " 
                << numSteps << ", " << 100. * i / numSteps 
                << " percent complete." << std::endl;
        }
        if ( !doIntegers ) // use the normally distributed numbers
        {
            // get random numbers
            gBrownian->calculateForce( );
            // save the random numbers
            gBrownian->saveState( fileName, i != 0 /* append if not the first step */ );
        }
        else // get the integers
        {
            // get random numbers
            randomGenerator->generateIntegers( );
            // save them
            if (!randomGenerator->saveIntegers( fileName, i != 0 )) {
		return false;
	    }
        }
    }
    std::cout << "Finished profiling random numbers, saved to " << fileName << std::endl;
    return true;
}

bool chromoSim::readState( std::vector<CS_REAL3> &pos, std::vector<CS_REAL3> &vel )
{
    if (!doing_okay()) { return false; }
    solver->readPosition( pos );
    solver->readVelocity( vel );
    return true;
}

// save current state
bool chromoSim::saveState( std::string fileName, bool append )
{
    // output position info
    std::ofstream outFile;
    if (append) {
        outFile.open( fileName, std::ios::out | std::ios::app );
    } else {
        outFile.open( fileName, std::ios::out );
    }
    // check if open
    if (!outFile.is_open()) {
        std::cerr << "ERROR: Unable to open "
            << fileName << "." << std::endl;
	return false;
    }

    // if we are appending, add an extra line
    if (append) {
        outFile << "\n";
    }
    
    // save our data
    outFile << solver->printPosition( getCurrentTime( ) );

    outFile.flush();
    outFile.close();
    return true;
}

// print information about the forces
bool chromoSim::printForces( )
{
    if (!doing_okay()) { return false; }
    std::cout << gDamping->printForce( ) << std::endl;
    std::cout << gSpring->printForce( ) << std::endl;
    std::cout << gHinge->printForce( ) << std::endl;
    std::cout << gBrownian->printForce( ) << std::endl;
    std::cout << gCollision->printForce( ) << std::endl;
    std::cout << solver->printVelocity( ) << std::endl;
    std::cout << solver->printPosition( getCurrentTime( ) ) << std::endl;
    return true;
}

// check NAN
bool chromoSim::checkNAN( )
{
    bool retval = false;

    if (gDamping->anyNAN())
    {
        std::cerr << "damping NAN" << std::endl;
        return true;
    }
    if (gSpring->anyNAN())
    {
        std::cerr << "springs NAN" << std::endl;
        return true;
    }
    if (gHinge->anyNAN())
    {
        std::cerr << "hinges NAN" << std::endl;
        return true;
    }
    if (gBrownian->anyNAN())
    {
        std::cerr << "brownian NAN" << std::endl;
        return true;
    }
    if (gCollision->anyNAN())
    {
        std::cerr << "collision NAN" << std::endl;
        return true;
    }
    if (solver->anyNAN())
    {
        std::cerr << "solver NAN" << std::endl;
        return true;
    }

    return false;
}

bool chromoSim::profileSimulation( int numSteps )
{
    // maximum values for each of the forces
    CS_REAL maxDamping = 0;
    CS_REAL maxSpring = 0;
    CS_REAL maxHinge = 0;
    CS_REAL maxBrownian = 0;
    CS_REAL maxCollision = 0;
    CS_REAL maxForce = 0;
    // maximum values for the velocities
    CS_REAL maxVelocity = 0;
    // count number of times we have a collision
    int timesCollision = 0;
    // simulate for that many time steps and get the velocities
    for ( int i = 0; i < numSteps; ++i ) {
        // do a time step
        simulateTimeStep( );

        // magic number, but for sort of debugging purposes... if <10 be more verbose
        if (numSteps < 10) {
            std::cout << "Step " << i+1 << "...\n";
            printForces( );
            std::cout << std::endl;
        }

        // check for NANs
        if (checkNAN( )) {
            std::cout << "NAN detected!" << std::endl;
	    return false;
        }

        // get maximums of forces
        CS_REAL maxf = 0;
        // damping
        maxf = gDamping->maximumMagnitude();
        if (maxf > maxDamping)
            maxDamping = maxf;
        // springs
        maxf = gSpring->maximumMagnitude();
        if (maxf > maxSpring)
            maxSpring = maxf;
        // hinges
        maxf = gHinge->maximumMagnitude();
        if (maxf > maxHinge)
            maxHinge = maxf;
        // brownian
        maxf = gBrownian->maximumMagnitude();
        if (maxf > maxBrownian)
            maxBrownian = maxf;
        // collisions
        maxf = gCollision->maximumMagnitude();
        if (maxf > maxCollision)
            maxCollision = maxf;
        if (maxf > 0) 
            ++timesCollision;
        // all the forces
        maxf = totalForce->maximumMagnitude();
        if (maxf > maxForce)
            maxForce = maxf;

        // get maximum of the velocity
        CS_REAL maxv = solver->maxVelocity();
        if (maxv > maxVelocity)
            maxVelocity = maxv;
    }
    
    std::cout << "Performed " << numSteps << " time steps of " << simTimeStep << " units each" << std::endl;
    std::cout << std::endl;
    std::cout << "Maximum forces reached in " << numSteps << " time steps:" << std::endl;
    std::cout << "Damping: " << maxDamping << std::endl;
    std::cout << "Spring: " << maxSpring << std::endl;
    std::cout << "Hinge: " << maxHinge << std::endl;
    std::cout << "Collision: " << maxCollision << std::endl;
    std::cout << "Brownian: " << maxBrownian << std::endl;
    std::cout << "Total: " << maxForce << std::endl;
    std::cout << std::endl;
    std::cout << "Maximum velocity reached: " << maxVelocity << std::endl;
    std::cout << "This corresponds to a maximum displacement of " << maxVelocity * simTimeStep << std::endl;
    std::cout << std::endl;
    std::cout << "Time steps with collisions: " << timesCollision 
        << " (" << (100. * timesCollision) / numSteps << " percent of time steps)" << std::endl;
    return true;
}

// destructor
chromoSim::~chromoSim( )
{
    // end operations occuring in the queue
    clFlush( queue );
    clFinish( queue );
    
    // delete our forces
    delete gDamping;
    delete gBrownian;
    delete gSpring;
    delete gHinge;
    delete gCollision;

    // delete our sum forces
    delete totalForce;

    // delete our solver
    delete solver;

    // release the command queue
    clReleaseCommandQueue( queue );
    // release the context
    clReleaseContext( context );
}

// constructor
chromoSim::chromoSim( MASS_node * mh, SPRING_node * sh, HINGE_node * hh,
        CS_REAL randomForce,
        CS_REAL springDamping, // the SOFA-esque damping constant associated with relative velocity over springs. F = -k * springDamping * relativeVelocity
        CS_REAL collisionSpringConstant,
        CS_REAL startingTime, // the time that we start off as saying we're at
        CS_REAL timeStep, // the time step used for each iteration
        int collisionScheme, // collision scheme to use -1 no collisions, 0 brute force spheres, 1 GPUGems spherees, 2 NVidia spheres, TODO more kinds
        bool useGPU, // asks openCL for a GPU if true (default), asks for a CPU if false
        int randomSeed, // seed being used by random number generator
        std::string clFilesDir // the directory to use to load OpenCL kernels
	, cl_ulong maximumInteractions  // Maximum particles interacting a single particle in any step.
        , int whichDevice   // Which device to attempt to open?
        )
{
    // We're not broken yet...
    broken = false;

    verbose = false;

    cl_ulong simTotalMaxInteractions = maximumInteractions;

    // get information about the time
    initialTime = startingTime;
    simTimeStep = timeStep;
    timeStepsMade = 0; // we start out having made no steps

    // information about the masses in the simulation
    // to be used in setting up solver and forces
    cl_ulong numParticles; // the number of particles
    CS_REAL3 * position; // the positions
    CS_REAL * mass; // the masses
    CS_REAL * massDamping; // the damping (drag = mass * massDamping, F = -drag * v)
    CS_REAL * massRadius; // the mass radii

    // acquire these parameters
    numParticles = 0;
    // get node to iterate over the masses
    MASS_node * m_it = mh;
    while (m_it != NULL) {
        m_it->sofaID = numParticles; // set the ID based off a zero index
        m_it = m_it->next;
        numParticles++;
    }
    // we have successfully found the total number of masses.
    // initialize an array for storing the positions and masses...
    position = new CS_REAL3[numParticles];
    mass = new CS_REAL[numParticles];
    massDamping = new CS_REAL[numParticles];
    massRadius = new CS_REAL[numParticles];
    // loop over the masses and give them values.
    m_it = mh; // go back to the beginning
    for (cl_ulong i = 0; i < numParticles; ++i) {
        // Note that we should have that i equals m_it->sofaID. If not, bad luck on us...
        // populate the position
        position[i].s[0] = static_cast<CS_REAL>(m_it->x);
        position[i].s[1] = static_cast<CS_REAL>(m_it->y);
        position[i].s[2] = static_cast<CS_REAL>(m_it->z);
        // populate the mass array
        mass[i] = static_cast<CS_REAL>( m_it->mass );
        // populate the mass radius array
        massRadius[i] = static_cast<CS_REAL>( m_it->radius );
        // give the mass its associated damping constant
        massDamping[i] = static_cast<CS_REAL>(m_it->damping);
        // next iteration, please...
        m_it = m_it->next;
    }
    // we now have details about the parameters of the masses

    // initialize OpenCL
    if (!initializeOpenCL( useGPU, whichDevice )) {
        std::cerr << "chromoSim::chromoSim(): Could not open requested device" << endl;
	broken = true;
	return;
    }
    
    // load our solver
    /*solver = new solverIMEXEuler( clFilesDir, context, 
        queue, simTimeStep, position, numParticles, 
        deviceId, mass, massDamping );*/
    solver = new solverImplicitEuler( clFilesDir, context,
        queue, simTimeStep, position, numParticles,
        deviceId, mass );
    if (!solver || !solver->doing_okay()) {
	broken = true;
	return;
    }

    // get position and velocity buffers for possible use with other forces
    cl_mem clPosition = solver->getCLPosition( );
    cl_mem clVelocity = solver->getCLVelocity( );

    // load damping forces
    gDamping = new forceMassDamping( clFilesDir, context, queue,
        numParticles, simTotalMaxInteractions, deviceId,
        clVelocity, mass, massDamping );
    // load Brownian forces
    // It is faster to do the random-number generation on the GPU using a
    // parallel generator than it is to generate the numbers on the CPU and
    // send them down.
    gBrownian = new forceRandomGPU( clFilesDir, context, queue, 
        numParticles, simTotalMaxInteractions, simTimeStep,
        deviceId, randomForce, randomSeed );
//    gBrownian = new forceRandomCPU( clFilesDir, context, queue, 
//        numParticles, simTotalMaxInteractions, simTimeStep,
//        randomForce, randomSeed );
    // load spring forces
    gSpring = new forceSpringsCL( clFilesDir, context, queue, 
        numParticles, simTotalMaxInteractions, deviceId, clPosition, clVelocity,
        sh, springDamping );
    // load hinge forces
    gHinge = new forceHingesCL( clFilesDir, context, queue, 
        numParticles, simTotalMaxInteractions, deviceId, clPosition, hh );
    
    // load collision forces
    if ( collisionScheme == -1 ) // there is no collision force
    {
        gCollision = new genericForce( clFilesDir, context, queue, 
            numParticles, simTotalMaxInteractions, false );
    }
    else if (collisionScheme == 0) // brute force (spheres)
    {
        gCollision = new forceCollisionsBruteSphere( clFilesDir, context, queue, numParticles,
            simTotalMaxInteractions, deviceId, clPosition, massRadius, collisionSpringConstant );
    }
    else if (collisionScheme == 1) // GPU Gems styled spatial subdivision (spheres)
    {
        gCollision = new forceCollisionsUniformSphere( clFilesDir, context, queue, numParticles,
            simTotalMaxInteractions, deviceId, clPosition, massRadius, collisionSpringConstant, 
            static_cast<CS_REAL>(1.5), static_cast<CS_REAL>(5), 3877 );
    }
    else if (collisionScheme == 2) // styled after NVidia OpenCL example (spheres)
    {
        gCollision = new forceCollisionsUniformSphere2( clFilesDir, context, queue, numParticles,
            simTotalMaxInteractions, platformId, deviceId, clPosition, massRadius, collisionSpringConstant, 
            static_cast<CS_REAL>(2.5), static_cast<cl_uint>(4 * numParticles + 1) );
    } else {
        std::cerr << "Collision scheme " << collisionScheme << " is undefined" << std::endl;
	broken = true;
	return;
    }

    // put these into a sumForces to compute and add them together when necessary
    std::vector<genericForce*> theForces;
    theForces.push_back(gDamping);
    theForces.push_back(gSpring);
    theForces.push_back(gCollision);
    theForces.push_back(gHinge);
    theForces.push_back(gBrownian);

    totalForce = new sumForces( clFilesDir, context, queue, numParticles, simTotalMaxInteractions,
        deviceId, theForces );

    // attach these forces to the integration scheme
    solver->attachForces( totalForce );

    // delete the temporary arrays we created
    delete[] position;
    delete[] mass;
    delete[] massDamping;
    delete[] massRadius;
}

bool chromoSim::initializeOpenCL( bool useGPU, unsigned whichDevice )
{
    // get platforms
    cl_uint platformIdCount = 0;
    clGetPlatformIDs( 0, NULL, &platformIdCount );
    if ( platformIdCount > 0 ) {
        if (verbose) std::cout << "Detected " << platformIdCount << " OpenCL platforms.\n";
    } else {
        std::cerr << "Did not detect any OpenCL platforms.\n";
        return false;
    }
    std::vector< cl_platform_id > platformIds( platformIdCount );
    clGetPlatformIDs( platformIdCount, platformIds.data(), NULL );
    // print out names of the platforms for informative purposes
    if (verbose) std::cout << "Platforms:" << std::endl;
    for (unsigned i = 0; i < platformIdCount; ++i) {
        // get size of the name
        size_t pNameSize = 0;
        clGetPlatformInfo( platformIds[i], CL_PLATFORM_NAME, 
            0, NULL, &pNameSize );
        // make string to read in the name
        std::string pName( pNameSize, '0' );
        clGetPlatformInfo( platformIds[i], CL_PLATFORM_NAME,
            pNameSize, (void *) pName.data(), NULL );
        if (verbose) std::cout << i << " : " << pName << std::endl;
    }
    // TODO: allow user to select which platform to use???
    if (verbose) std::cout << "\nUsing platform 0...\n";
    platformId = platformIds[0];

    // get devices
    cl_uint deviceIdCount = 0;
    cl_device_type myDeviceType = 
        useGPU ? CL_DEVICE_TYPE_GPU | CL_DEVICE_TYPE_ACCELERATOR : CL_DEVICE_TYPE_CPU;
    clGetDeviceIDs( platformId, myDeviceType, 0, NULL, &deviceIdCount );
    if ( deviceIdCount > 0 ) {
        if (verbose) std::cout << "Detected " << deviceIdCount << " devices.\n";
    } else {
        std::cerr << "Did not detect any devices.\n";
        return false;
    }
    std::vector< cl_device_id > deviceIds( deviceIdCount );
    clGetDeviceIDs( platformId, myDeviceType, deviceIdCount, deviceIds.data(), NULL );
    // print out names of the devices for informative purposes
    if (verbose) std::cout << "Devices: " << std::endl;
    for (unsigned i = 0; i < deviceIdCount; ++i) {
        cl_uint addressBits;	// OpenCL number of address bits
        size_t maxGroupSize;	// OpenCL maximum size of a kernel group (XXX size_t varies)
        size_t maxWorkItemSizes[3];	// OpenCL maximum work item sizes (XXX size_t varies)
        size_t dNameSize = 0;
        clGetDeviceInfo( deviceIds[i], CL_DEVICE_NAME, 0, NULL, &dNameSize );
        std::string dName( dNameSize, '0' );
        clGetDeviceInfo( deviceIds[i], CL_DEVICE_NAME, dNameSize,
            (void *) dName.data(), NULL );
        if (verbose) std::cout << i << " : " << dName << std::endl;

        clGetDeviceInfo( deviceIds[i], CL_DEVICE_ADDRESS_BITS, sizeof(addressBits), &addressBits, NULL );
        if (verbose) std::cout << "  : (" << addressBits << " address bits)" << std::endl;
        clGetDeviceInfo( deviceIds[i], CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(maxGroupSize), &maxGroupSize, NULL );
        if (verbose) std::cout << "  : (" << maxGroupSize << " max work-group size)" << std::endl;
        clGetDeviceInfo( deviceIds[i], CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(maxWorkItemSizes), maxWorkItemSizes, NULL );
        if (verbose) std::cout << "  : (" << maxWorkItemSizes[0] << "x" << maxWorkItemSizes[1] << "x" << maxWorkItemSizes[2] << " max work-item sizes)" << std::endl;
    }
    if (whichDevice >= deviceIds.size()) {
        std::cerr << "chromoSim::initializeOpenCL(): Requested nonexistent device" << endl;
        return false;
    }
    if (verbose) std::cout << "\nUsing device " << whichDevice << "...\n";
    deviceId = deviceIds[whichDevice];

    // use the device to get a context
    cl_int error;
    context = clCreateContext( NULL, // XXX: could do something with contextProperties, but didn't seem relevant...
        1, &deviceId,
        NULL, NULL, // something about callbacks, seems unimportant
        &error );
    if (CheckErr( error, "clCreateContext" )) {
	return false;
    }
    if (verbose) std::cout << "Generated OpenCL Context." << std::endl;
    

    // generate the command queue on the context and device
    queue = clCreateCommandQueue( context, deviceId, 0, // XXX: not worrying about command queue properties
        &error );
    if (CheckErr( error, "clCreateCommandQueue" )) {
	return false;
    }
    if (verbose) std::cout << "Generated OpenCL Command Queue." << std::endl;

    // that's it for the nonspecific stuff
    return true;
}

