// forceCollisionsBruteSphere.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "forceCollisionsBruteSphere.h"

// what's the name of this force?
const char * forceCollisionsBruteSphere::getName( )
{
    return "Force BF Spherical Collision";
}

// constructor
forceCollisionsBruteSphere::forceCollisionsBruteSphere(
    std::string sim_CLdir, // the path of directory with openCL source
    cl_context sim_context, // openCL context to use
    cl_command_queue sim_queue, // openCL queue to use
    cl_ulong sim_numParticles, // number of particles in the simulation
    cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
    cl_device_id sim_deviceId, // the deviceId we are running on
    cl_mem sim_clPosition, // the positions
    CS_REAL * sim_MassRadius, // the masses (radius information)
    CS_REAL collisionSpringConstant
    ) : genericForce( sim_CLdir, sim_context, sim_queue, sim_numParticles, sim_totalMaxInteractions, true )
{
    // set collisionConstant
    collisionConstant = collisionSpringConstant;
    
    // only one thread will access a given force at a time
    columns = 1;
    // create a force vector and set it to zero on host and device
    initForcesZero( );

    // we will use as much space as is allowed by totalMaxInteractions for collisions
    jacobianMaxInteractions = totalMaxInteractions;
    // set up the buffers for the Jacobian
    initJacobianZero( );
    // did we successfully allocate a Jacobian? if so, set some default values for buffer
    if (clJacobianLoaded)
    {
        // set the diagonal of the Jacobian to the first numParticles indices
        // set the current interaction count to 1
        cl_ulong * interactionCount = new cl_ulong[numParticles];
        cl_ulong * interactionIndex = new cl_ulong[numParticles*jacobianMaxInteractions];
        for (cl_ulong i = 0; i < numParticles; ++i)
        {
            interactionCount[i] = 1;
            interactionIndex[i] = i;
            for (cl_ulong j = 1; j < jacobianMaxInteractions; ++j)
            {
                interactionIndex[j*numParticles + i] = 0; // they aren't really interacting with 0, but for the sake of having it be defined
            }
        }
        checkErr( clEnqueueWriteBuffer( queue, clJacobian.numInteractions, CL_TRUE,
            0, numParticles * sizeof(cl_ulong), interactionCount, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (numInteractions)" );
        checkErr( clEnqueueWriteBuffer( queue, clJacobian.interactionParticleIndex, CL_TRUE,
            0, numParticles * jacobianMaxInteractions * sizeof(cl_ulong),
            interactionIndex, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (interactionParticleIndex)" );

        delete[] interactionCount;
        delete[] interactionIndex;
    }

    // load the program "forceCollisionsBruteSphere.cl"
    loadProgram( sim_deviceId, "forceCollisionsBruteSphere.cl" );

    cl_int error = 0;
    
    // copy the mass radius array to device
    clMassRadius = clCreateBuffer( context, // put into our class' context
        CL_MEM_READ_ONLY, // make the buffer read only
        numParticles * sizeof( CS_REAL ), // allocate space for the springs
        NULL, &error);
    checkErr( error, "clCreateBuffer" );
    checkErr(
        clEnqueueWriteBuffer( queue, clMassRadius, CL_TRUE, 0,
            numParticles * sizeof( CS_REAL ), sim_MassRadius,
            0, NULL, NULL),
        "clEnqueueWriteBuffer"
        );

    // load kernel from program
    clKernelCollision = clCreateKernel( program,
        "applySphericalCollisions", &error );
    checkErr( error, "clCreateKernel" );

    // pass arguments into kernel
    error = clSetKernelArg( clKernelCollision, 0, 
        sizeof(CS_REAL), &collisionConstant );
    error |= clSetKernelArg( clKernelCollision, 1,
        sizeof(cl_ulong), &numParticles );
    error |= clSetKernelArg( clKernelCollision, 2,
        sizeof(cl_mem), &clMassRadius );
    error |= clSetKernelArg( clKernelCollision, 3,
        sizeof(cl_mem), &sim_clPosition );
    error |= clSetKernelArg( clKernelCollision, 4,
        sizeof(cl_mem), &clForce );
    checkErr( error, "clSetKernelArg" );
    
    // get corrected global size for clKernelCollision
    clKernelCollisionGlobalSize = getKernelPaddedGlobalSize( 
        sim_deviceId, clKernelCollision, numParticles );

    if (clJacobianLoaded) // load kernel for computing Jacobian
    {
        clKernelCollisionJacobian = clCreateKernel( program,
            "applySphericalCollisionsAndJacobian", &error );
        checkErr( error, "clCreateKernel (clKernelCollisionJacobian)" );

        // pass arguments into kernel
        error = clSetKernelArg( clKernelCollisionJacobian, 0, 
            sizeof(CS_REAL), &collisionConstant );
        error |= clSetKernelArg( clKernelCollisionJacobian, 1,
            sizeof(cl_ulong), &numParticles );
        error |= clSetKernelArg( clKernelCollisionJacobian, 2,
            sizeof(cl_mem), &clMassRadius );
        error |= clSetKernelArg( clKernelCollisionJacobian, 3,
            sizeof(cl_mem), &sim_clPosition );
        error |= clSetKernelArg( clKernelCollisionJacobian, 4,
            sizeof(cl_mem), &clForce );
        error |= clSetKernelArg( clKernelCollisionJacobian, 5,
            sizeof(cl_mem), &clJacobian.j_PosX );
        error |= clSetKernelArg( clKernelCollisionJacobian, 6,
            sizeof(cl_mem), &clJacobian.j_PosY );
        error |= clSetKernelArg( clKernelCollisionJacobian, 7,
            sizeof(cl_mem), &clJacobian.j_PosZ );
        error |= clSetKernelArg( clKernelCollisionJacobian, 8,
            sizeof(cl_mem), &clJacobian.interactionParticleIndex );
        error |= clSetKernelArg( clKernelCollisionJacobian, 9,
            sizeof(cl_mem), &clJacobian.numInteractions );
        checkErr( error, "clSetKernelArg (Jacobian)" );

        // get the global size to use
        clKernelCollisionJacobianGlobalSize = getKernelPaddedGlobalSize(
            sim_deviceId, clKernelCollisionJacobian, numParticles );
    }
}

// destructor
forceCollisionsBruteSphere::~forceCollisionsBruteSphere( )
{
    clReleaseKernel( clKernelCollision );
    clReleaseMemObject( clMassRadius );
}

// calculating the force
void forceCollisionsBruteSphere::calculateForce( )
{
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // we are adding it to our command queue
            clKernelCollision, // the kernel we are running (integrate)
            1, // we have a 1 dimensional work dimension
            NULL, // we aren't using any offsets for the work dimensions
            &clKernelCollisionGlobalSize, // our global work space is the number of particles
            NULL, // local item size chosen by implementation
            0, NULL, NULL // we aren't waiting for any events
            ),
        "clEnqueueNDRangeKernel");

    // the host and device arrays are no longer in synchronization
    updated = false;
}

// calculate force and Jacobian
void forceCollisionsBruteSphere::calculateForceAndJacobian( )
{
    if (!clJacobianLoaded)
    {
        // just calculate the force
        calculateForce( );
    }
    else
    {
        checkErr(
            clEnqueueNDRangeKernel( // run the kernel
                queue, // we are adding it to our command queue
                clKernelCollisionJacobian, // the kernel we are running (integrate)
                1, // we have a 1 dimensional work dimension
                NULL, // we aren't using any offsets for the work dimensions
                &clKernelCollisionJacobianGlobalSize, // our global work space is the number of particles
                NULL, // local item size chosen by implementation
                0, NULL, NULL // we aren't waiting for any events
                ),
            "clEnqueueNDRangeKernel (Jacobian)");

        // the host and device arrays are no longer in synchronization
        updated = false;
    }
}
