// forceCollisionsUniformSphere2.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "forceCollisionsUniformSphere2.h"
#include <algorithm> // for stable_sort

#include <iostream>
#include <cstdlib>
#include <cmath>
#include <vector> // while picking grid scale
#include <algorithm> // for sorting in picking grid scale

// what's the name of this force?
const char * forceCollisionsUniformSphere2::getName( )
{
    return "Force UVSD2 Spherical Collision";
}

// static array with good grid scales. 20 chosen prime numbers
const cl_uint forceCollisionsUniformSphere2::lookupGridScale[] 
= {3877,3881,3889,3907,4021,4027,4049,4051,4057,4073, 
    4789,4793,5503,5519,5521,5527,5531,5557,5563,5569};

// static function for getting a grid scale that doesn't cause adjacent
// grid coordinates as calculated in the CL file to be mapped to the same hash.
cl_int forceCollisionsUniformSphere2::pickGridScale( cl_uint sim_maxHash )
{
    // loop over the 20 values in the lookup table
    for (int i = 0; i < 20; ++i)
    {
        // get the possible grid scale
        cl_uint gs = lookupGridScale[i];
        // get the values that this grid scale maps the "origin"
        // and the 26 other adjacent cells to.
        std::vector<cl_uint> values;
        for (int dx = -1; dx <= 1; ++dx)
            for (int dy = -1; dy <= 1; ++dy)
                for (int dz = -1; dz <= 1; ++dz)
                    values.push_back( (dx + gs*(dy + gs*dz))%sim_maxHash );
        // sort the values
        std::sort( values.begin(), values.end() );
        // check for repeats
        bool noRepeats = std::adjacent_find( values.begin(), values.end() ) == values.end();
        // if no repeats, use this
        if (noRepeats)
        {
            return gs;
        }
        // otherwise, on to the next possible value.
    }
    // by numerical experiment with the above look up table, if maxHash=4*N + 1,
    // we only get these problems with adjacent hashes when N=1,2,3,4,5,6,8, (trying N up to 500,000).
    // therefore, we can be fairly satisfied that, along with a restriction on the minimum hash value,
    // we will not have problems with double counting in our force calculations code. However, if we get
    // here, we should not continue.
    std::cerr << "Error: Could not pick adequate grid scale in forceCollisionsUniformSphere2." << std::endl;
    return -1;
}

// constructor
forceCollisionsUniformSphere2::forceCollisionsUniformSphere2(
    std::string sim_CLdir, // the path of directory with openCL source
    cl_context sim_context, // openCL context to use
    cl_command_queue sim_queue, // openCL queue to use
    cl_ulong sim_numParticles, // number of particles in the simulation
    cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
    cl_platform_id sim_platformId, // the platformId we are running on
    cl_device_id sim_deviceId, // the deviceId we are running on
    cl_mem sim_clPosition, // the positions
    CS_REAL * sim_MassRadius, // the mass radius information
    CS_REAL collisionSpringConstant,
    CS_REAL cellDimensionScale, // the proportion of cell length size to the largest mass radius (~ 2.5)
    cl_uint sim_maxHash // maximum value of hash function
    ) : genericForce( sim_CLdir, sim_context, sim_queue, sim_numParticles, sim_totalMaxInteractions, true )
{
    // set collisionConstant
    collisionConstant = collisionSpringConstant;

    // set maxHash
    maxHash = sim_maxHash;
    // override smaller sizes.
    if (maxHash < 1049) // arbitrarily chosen prime number over 1000
    {
        maxHash = 1049;
    }
    cl_int gridScale = forceCollisionsUniformSphere2::pickGridScale( maxHash );
    if (gridScale < 0) {
	cerr << "forceCollisionsUniformSphere2::forceCollisionsUniformSphere2(): Invalid grid scale" << endl;
	return;
    }
    
    // only one thread for each particle calculating collision forces
    columns = 1;

    // variable for getting errors from OpenCL API
    cl_int error = 0;

    // copy the mass radius array to device
    clMassRadius = clCreateBuffer( context, // put into our class' context
        CL_MEM_READ_ONLY, // make the buffer read only
        numParticles * sizeof( CS_REAL ), // allocate space for the radii
        NULL, &error);
    checkErr( error, "clCreateBuffer (massRadius)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clMassRadius, CL_TRUE, 0,
            numParticles * sizeof( CS_REAL ), sim_MassRadius,
            0, NULL, NULL),
        "clEnqueueWriteBuffer (massRadius)"
        );

    // get the maximum massRadius
    CS_REAL maxMassRadius = 0;
    for (cl_ulong i = 0; i < numParticles; ++i) {
        CS_REAL mr = sim_MassRadius[i];
        // check if the mass radius is larger than the current known maximum
        if (mr > maxMassRadius) {
            maxMassRadius = mr;
        }
    }

    // create structure for hashes of positions in the grid, send to device
    positionHash = new cl_uint2[numParticles];
    for (cl_uint i = 0; i < numParticles; ++i)
    {
        positionHash[i].s[0] = 0;
        positionHash[i].s[1] = i;
    }
    clPositionHash = clCreateBuffer( context, CL_MEM_READ_WRITE,
        numParticles * sizeof( cl_uint2 ), NULL, &error);
    checkErr( error, "clCreateBuffer (positionHash)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clPositionHash, CL_TRUE, 0,
            numParticles * sizeof( cl_uint2 ), positionHash,
            0, NULL, NULL),
        "clEnqueueWriteBuffer (positionHash)"
        );

    // create emptyCollisionCell for clCollisionCellStart and End
    emptyCollisionCell = new cl_uint[maxHash];
    for (cl_ulong i = 0; i < maxHash; ++i)
    {
        emptyCollisionCell[i] = -1;
    }

    // initialize clCollisionCellStart and End
    clCollisionCellStart = clCreateBuffer( context, CL_MEM_READ_WRITE,
        maxHash * sizeof( cl_uint ), NULL, &error);
    checkErr( error, "clCreateBuffer (clCollisionCellStart)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clCollisionCellStart, CL_TRUE, 0,
            maxHash * sizeof( cl_uint ), emptyCollisionCell,
            0, NULL, NULL ),
        "clEnqueueWriteBuffer (clCollisionCellStart)"
        );
    clCollisionCellEnd = clCreateBuffer( context, CL_MEM_READ_WRITE,
        maxHash * sizeof( cl_uint ), NULL, &error);
    checkErr( error, "clCreateBuffer (clCollisionCellEnd)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clCollisionCellEnd, CL_TRUE, 0,
            maxHash * sizeof( cl_uint ), emptyCollisionCell,
            0, NULL, NULL ),
        "clEnqueueWriteBuffer (clCollisionCellEnd)"
        );

    // create a force vector and set it to zero on host and device
    initForcesZero( );

    // we will use as much space as is allowed by totalMaxInteractions for collisions
    jacobianMaxInteractions = totalMaxInteractions;
    // set up the buffers for the Jacobian
    initJacobianZero( );
    // did we successfully allocate a Jacobian? if so, set some default values for buffer
    if (clJacobianLoaded)
    {
        // set the diagonal of the Jacobian to the first numParticles indices
        // set the current interaction count to 1
        cl_ulong * interactionCount = new cl_ulong[numParticles];
        if (!interactionCount) {
            fprintf(stderr,"forceCollisionsUniformSphere2::forceCollisionsUniformSphere2(): Out of memory allocating interactionCount\n");
            exit(-1);
        }
        cl_ulong * interactionIndex = new cl_ulong[numParticles*jacobianMaxInteractions];
        if (!interactionIndex) {
            fprintf(stderr,"forceCollisionsUniformSphere2::forceCollisionsUniformSphere2(): Out of memory allocating interactionIndex\n");
            exit(-1);
        }
        for (cl_ulong i = 0; i < numParticles; ++i)
        {
            interactionCount[i] = 1;
            interactionIndex[i] = i;
            for (cl_ulong j = 1; j < jacobianMaxInteractions; ++j)
            {
                interactionIndex[j*numParticles + i] = 0;
            }
        }      
        checkErr( clEnqueueWriteBuffer( queue, clJacobian.numInteractions, CL_TRUE,
            0, numParticles * sizeof(cl_ulong), interactionCount, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (numInteractions)" );
        checkErr( clEnqueueWriteBuffer( queue, clJacobian.interactionParticleIndex, CL_TRUE,
            0, numParticles * jacobianMaxInteractions * sizeof(cl_ulong),
            interactionIndex, 0, NULL, NULL ),
            "clEnqueueWriteBuffer (interactionParticleIndex)" );

        // Wait for the buffers to be sent before deleting the arrays we are sending.
        checkErr( clFlush( queue ), "clFlush (after clEnqueueWriteBuffer)" );
        checkErr( clFinish( queue ), "clFinish (after clEnqueueWriteBuffer)" );
        delete[] interactionCount;
        delete[] interactionIndex;
    }

    // create our grid scale and the cell dimensions
    cl_uint3 clGridScale;
    clGridScale.s[0] = 1;
    clGridScale.s[1] = static_cast<cl_uint>(gridScale);
    clGridScale.s[2] = static_cast<cl_uint>(gridScale * gridScale);
    CS_REAL3 clCellDimensions;
    clCellDimensions.s[0] = maxMassRadius * cellDimensionScale;
    clCellDimensions.s[1] = maxMassRadius * cellDimensionScale;
    clCellDimensions.s[2] = maxMassRadius * cellDimensionScale;

    // load the program "forceCollisionsUniformSphere2.cl"
    loadProgram( sim_deviceId, "forceCollisionsUniformSphere2.cl" );

    // load the kernel to get the hashes
    clKernelGetHash = clCreateKernel( program,
        "calculateHashedPosition", &error );
    checkErr( error, "clCreateKernel (clKernelGetHash)" );
    // set arguments to this kernel
    error = clSetKernelArg( clKernelGetHash, 0, 
        sizeof( cl_uint3 ), &clGridScale );
    error |= clSetKernelArg( clKernelGetHash, 1,
        sizeof( CS_REAL3 ), &clCellDimensions );
    error |= clSetKernelArg( clKernelGetHash, 2,
        sizeof( cl_uint ), &maxHash );
    error |= clSetKernelArg( clKernelGetHash, 3,
        sizeof( cl_ulong ), &numParticles );
    error |= clSetKernelArg( clKernelGetHash, 4,
        sizeof( cl_mem ), &sim_clPosition );
    error |= clSetKernelArg( clKernelGetHash, 5,
        sizeof( cl_mem ), &clPositionHash );
    checkErr( error, "clSetKernelArg (clKernelGetHash)" );

    // get global work item size for this
    clKernelGetHashGlobalSize = getKernelPaddedGlobalSize( 
        sim_deviceId, clKernelGetHash, numParticles );

    // sort using clpp
    // set up directory to get clpp OpenCL kernels from
    std::ostringstream path;
    path << CLdir << "/clpp/";
    clppProgram::setBasePath( path.str() );
    // set up clpp context
    sortContext.setup( sim_platformId, sim_deviceId, sim_context, sim_queue );
    // set up the sort routine
    // get the total number of bits necessary to represent any of the hash values
    unsigned int bitsToSort = static_cast<unsigned int>( log( static_cast<double>(maxHash) ) / log( 2. )  ) + 1;
    // choose routine based on if device is CPU or GPU
    if (sortContext.isCPU)
    {
        sortRoutine = new clppSort_RadixSort( &sortContext, static_cast<unsigned>(numParticles), bitsToSort, false );
    }
    else // isGPU
    {
        sortRoutine = new clppSort_RadixSortGPU( &sortContext, static_cast<unsigned>(numParticles), bitsToSort, false );
    }
    sortRoutine->pushCLDatas( clPositionHash, numParticles );

    // load the kernel to finding cell bounds
    clKernelGetBounds = clCreateKernel( program,
        "findCellBounds", &error );
    checkErr( error, "clCreateKernel (clKernelGetBounds)" );
    // set arguments to this kernel
    error = clSetKernelArg( clKernelGetBounds, 0,
        sizeof( cl_mem ), &clPositionHash );
    error |= clSetKernelArg( clKernelGetBounds, 1,
        sizeof( cl_ulong ), &numParticles );
    // set third argument (index 2, of course) after finding local size
    error |= clSetKernelArg( clKernelGetBounds, 3,
        sizeof( cl_mem ), &clCollisionCellStart );
    error |= clSetKernelArg( clKernelGetBounds, 4,
        sizeof( cl_mem ), &clCollisionCellEnd );
    checkErr( error, "clSetKernelArg (clKernelGetBounds, nonlocal)" );
    // find global and local size
    clKernelGetBoundsGlobalSize = getKernelPaddedGlobalSize( 
        sim_deviceId, clKernelGetBounds, numParticles );
    clKernelGetBoundsLocalSize = getKernelMinimumPreferredLocalSize(
        sim_deviceId, clKernelGetBounds);
    error = clSetKernelArg( clKernelGetBounds, 2,
        (clKernelGetBoundsLocalSize + 1) * sizeof(cl_uint), NULL );
    checkErr( error, "clSetKernelArg (clKernelGetBounds, local)" );

    // load the kernel to compute the forces
    clKernelCollision = clCreateKernel( program,
        "applyCollisions", &error );
    checkErr( error, "clCreateKernel (clKernelCollision)" );
    // set arguments to this kernel
    error = clSetKernelArg( clKernelCollision, 0, 
        sizeof(CS_REAL), &collisionConstant );
    error |= clSetKernelArg( clKernelCollision, 1,
        sizeof( cl_uint3 ), &clGridScale );
    error |= clSetKernelArg( clKernelCollision, 2,
        sizeof( CS_REAL3 ), &clCellDimensions );
    error |= clSetKernelArg( clKernelCollision, 3,
        sizeof( cl_uint ), &maxHash );
    error |= clSetKernelArg( clKernelCollision, 4,
        sizeof( cl_ulong ), &numParticles );
    error |= clSetKernelArg( clKernelCollision, 5,
        sizeof( cl_mem ), &clCollisionCellStart );
    error |= clSetKernelArg( clKernelCollision, 6,
        sizeof( cl_mem ), &clCollisionCellEnd );
    error |= clSetKernelArg( clKernelCollision, 7,
        sizeof(cl_mem), &clPositionHash );
    error |= clSetKernelArg( clKernelCollision, 8,
        sizeof(cl_mem), &clMassRadius );
    error |= clSetKernelArg( clKernelCollision, 9,
        sizeof(cl_mem), &sim_clPosition );
    error |= clSetKernelArg( clKernelCollision, 10,
        sizeof(cl_mem), &clForce );
    checkErr( error, "clSetKernelArg (clKernelCollision)" );

    // get corrected global size for clKernelCollision
    clKernelCollisionGlobalSize = getKernelPaddedGlobalSize( 
        sim_deviceId, clKernelCollision, numParticles );

    if (clJacobianLoaded) // load kernel for computing Jacobian
    {
        clKernelCollisionJacobian = clCreateKernel( program,
            "applyCollisionsAndJacobian", &error );
        checkErr( error, "clCreateKernel (clKernelCollisionJacobian)" );
        // set arguments to this kernel
        error = clSetKernelArg( clKernelCollisionJacobian, 0, 
            sizeof(CS_REAL), &collisionConstant );
        error |= clSetKernelArg( clKernelCollisionJacobian, 1,
            sizeof( cl_uint3 ), &clGridScale );
        error |= clSetKernelArg( clKernelCollisionJacobian, 2,
            sizeof( CS_REAL3 ), &clCellDimensions );
        error |= clSetKernelArg( clKernelCollisionJacobian, 3,
            sizeof( cl_uint ), &maxHash );
        error |= clSetKernelArg( clKernelCollisionJacobian, 4,
            sizeof( cl_ulong ), &numParticles );
        error |= clSetKernelArg( clKernelCollisionJacobian, 5,
            sizeof( cl_mem ), &clCollisionCellStart );
        error |= clSetKernelArg( clKernelCollisionJacobian, 6,
            sizeof( cl_mem ), &clCollisionCellEnd );
        error |= clSetKernelArg( clKernelCollisionJacobian, 7,
            sizeof(cl_mem), &clPositionHash );
        error |= clSetKernelArg( clKernelCollisionJacobian, 8,
            sizeof(cl_mem), &clMassRadius );
        error |= clSetKernelArg( clKernelCollisionJacobian, 9,
            sizeof(cl_mem), &sim_clPosition );
        error |= clSetKernelArg( clKernelCollisionJacobian, 10,
            sizeof(cl_mem), &clForce );
        error |= clSetKernelArg( clKernelCollisionJacobian, 11,
            sizeof(cl_mem), &clJacobian.j_PosX );
        error |= clSetKernelArg( clKernelCollisionJacobian, 12,
            sizeof(cl_mem), &clJacobian.j_PosY );
        error |= clSetKernelArg( clKernelCollisionJacobian, 13,
            sizeof(cl_mem), &clJacobian.j_PosZ );
        error |= clSetKernelArg( clKernelCollisionJacobian, 14,
            sizeof(cl_mem), &clJacobian.interactionParticleIndex );
        error |= clSetKernelArg( clKernelCollisionJacobian, 15,
            sizeof(cl_mem), &clJacobian.numInteractions );
        checkErr( error, "clSetKernelArg (clKernelCollisionJacobian)" );
        // TODO: When the kernel is done, need to check if there were
        // ever too many interactions for a given particle.

        // get corrected global size for clKernelCollisionJacobian
        clKernelCollisionJacobianGlobalSize = getKernelPaddedGlobalSize( 
            sim_deviceId, clKernelCollisionJacobian, numParticles );
    }
}

// destructor
forceCollisionsUniformSphere2::~forceCollisionsUniformSphere2( )
{
    delete[] positionHash;
    delete[] emptyCollisionCell;
    delete sortRoutine;
    
    clReleaseKernel( clKernelGetHash );
    clReleaseKernel( clKernelGetBounds );
    clReleaseKernel( clKernelCollision );
    clReleaseMemObject( clMassRadius );
    clReleaseMemObject( clPositionHash );
    clReleaseMemObject( clCollisionCellStart );
    clReleaseMemObject( clCollisionCellEnd );
}

// calculating the force
void forceCollisionsUniformSphere2::calculateForce( )
{
    helperCalculateForce( );

    // calculate the forces
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // we are adding it to our command queue
            clKernelCollision, // the kernel we are running (integrate)
            1, // we have a 1 dimensional work dimension
            NULL, // we aren't using any offsets for the work dimensions
            &clKernelCollisionGlobalSize, // our global work space is the number of particles
            NULL, // local item size chosen by implementation
            0, NULL, NULL // we aren't waiting for any events
            ),
        "clEnqueueNDRangeKernel (clKernelCollision)");

    // the host and device arrays are no longer in synchronization
    updated = false;
}

void forceCollisionsUniformSphere2::calculateForceAndJacobian( )
{
    if (!clJacobianLoaded)
    {
        // just calculate the force
        calculateForce( );
    }
    else
    {
        helperCalculateForce( );

        // calculate the forces and Jacobian
        checkErr(
            clEnqueueNDRangeKernel( // run the kernel
                queue, // we are adding it to our command queue
                clKernelCollisionJacobian, // the kernel we are running (integrate)
                1, // we have a 1 dimensional work dimension
                NULL, // we aren't using any offsets for the work dimensions
                &clKernelCollisionJacobianGlobalSize, // our global work space is the number of particles
                NULL, // local item size chosen by implementation
                0, NULL, NULL // we aren't waiting for any events
                ),
            "clEnqueueNDRangeKernel (clKernelCollisionJacobian)");

        // the host and device arrays are no longer in synchronization
        updated = false;
    }
}

void forceCollisionsUniformSphere2::helperCalculateForce( )
{
    cl_int error = 0;
    
    // get hash information from current positions
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // we are adding it to our command queue
            clKernelGetHash, // the kernel we are running (integrate)
            1, // we have a 1 dimensional work dimension
            NULL, // we aren't using any offsets for the work dimensions
            &clKernelGetHashGlobalSize, // our global work size
            NULL, // local item size chosen by implementation
            0, NULL, NULL // we aren't waiting for any events
            ),
        "clEnqueueNDRangeKernel (clKernelGetHash)");

    // sort the hash information with respect to hash
    sortRoutine->sort();
    sortRoutine->waitCompletion();

    // set clCollisionCellStart to empty value
    checkErr(
        clEnqueueWriteBuffer( queue, clCollisionCellStart, CL_TRUE, 0,
            maxHash * sizeof( cl_uint ), emptyCollisionCell,
            0, NULL, NULL ),
        "clEnqueueWriteBuffer (clCollisionCellStart)"
        );

    // update cell starts and ends
    checkErr(
        clEnqueueNDRangeKernel( // run on kernel
        queue, // the queue to run on
        clKernelGetBounds, // the kernel to use
        1, // one dimensional input
        NULL, // no offsets
        &clKernelGetBoundsGlobalSize, // global size
        &clKernelGetBoundsLocalSize, // local size
        0, NULL, NULL // no events used
        ),
        "clEnqueueNDRangeKernel (clKernelGetBounds)"
        );
}

bool forceCollisionsUniformSphere2::comparisonPOSITION_HASH( const cl_uint2 & left, 
    const cl_uint2 & right )
{
    return left.s[0] < right.s[0];
}
