// forceRandomGPU.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "forceRandomGPU.h"
// #include "Random123/threefry.h" (doesn't work on BASS)

// what's the name of this force?
const char * forceRandomGPU::getName( )
{
    return "Force Brownian (GPU)";
}

// constructor
forceRandomGPU::forceRandomGPU(
    std::string sim_CLdir, // the path of directory with our openCL source code
    cl_context sim_context, // OpenCL context passed in from simulation
    cl_command_queue sim_queue, // OpenCL queue passed in from simulation,
    cl_ulong sim_numParticles, // number of particles in the simulation
    cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
    CS_REAL sim_timeStep, // the time step in the simulation
    cl_device_id sim_deviceId, // the deviceId we are running on
    CS_REAL randomForce, // the random force parameter from the configuration file
    int randomSeed // seed being used by the random number generator
    ) : genericForce( sim_CLdir, sim_context, sim_queue, sim_numParticles, sim_totalMaxInteractions, true )
{
    // genericForce constructor has set up the queue, the number of particles, and the time step.

    // start setting theRandomForce by setting to randomForce from configuration.
    theRandomForce = randomForce;

    // set up the number of columns, and set the forces (initially) to zero
    columns = 1; // we only have one source for the random force
    initForcesZero( );

    // value for checking errors from OpenCL API calls
    cl_int error = 0;
    
    // seed the random engine and set it up
    cl_uint4 * key = new cl_uint4[numParticles];
    cl_uint4 * counter = new cl_uint4[numParticles];
    for (cl_ulong i = 0; i < numParticles; ++i)
    {
        // set the key that is the same for the entire simulation
        key[i].s[0] = static_cast<cl_uint>(i); // the index of the particle
        key[i].s[1] = randomSeed; // user defined seed
        key[i].s[2] = 0xfade1337; // two arbitrary 32bit values
        key[i].s[3] = 0xbeeff00d;
        // set the starting value for the counter in our simulation
        counter[i].s[0] = 0; // should be arbitrary so long as we increment appropriately
        counter[i].s[1] = 0xdecafbad; // three arbitrary 32bit values
        counter[i].s[2] = 0x19931993;
        counter[i].s[3] = 0x42424242;
    }
    // send these arrays to device
    clRandomKey = clCreateBuffer( context,
        CL_MEM_READ_ONLY, // make this buffer read only
        numParticles * sizeof( cl_uint4 ),
        NULL, &error);
    checkErr( error, "clCreateBuffer (clRandomKey)" );
    checkErr(
        clEnqueueWriteBuffer( queue, clRandomKey, CL_TRUE, 0,
            numParticles * sizeof( cl_uint4 ),
            key, 0, NULL, NULL ),
        "clEnqueueWriteBuffer (clRandomKey)" );
    clRandomCounter = clCreateBuffer( context,
        CL_MEM_READ_WRITE, // this buffer is read-write
        numParticles * sizeof( cl_uint4 ),
        NULL, &error);
    checkErr( error, "clCreateBuffer (clRandomCounter)" );
    checkErr( 
        clEnqueueWriteBuffer( queue, clRandomCounter, CL_TRUE, 0,
            numParticles * sizeof( cl_uint4 ),
            counter, 0, NULL, NULL ),
        "clEnqueueWriteBuffer (clRandomCounter)" );
    // release memory for key and counter on host
    delete[] key;
    delete[] counter;

    // load the program "forceRandomGPU.cl"
    loadProgram( sim_deviceId, "forceRandomGPU.cl" );

    // load the random force calculations kernel from the program
    clKernelBrownian = clCreateKernel( program, "getRandomForce", &error );
    checkErr( error, "clCreateKernel" );

    // pass arguments into kernel
    error = clSetKernelArg( clKernelBrownian, 0, 
        sizeof(cl_mem), &clRandomKey );
    error |= clSetKernelArg( clKernelBrownian, 1,
        sizeof(cl_mem), &clRandomCounter );
    error |= clSetKernelArg( clKernelBrownian, 2,
        sizeof(CS_REAL), &theRandomForce );
    error |= clSetKernelArg( clKernelBrownian, 3,
        sizeof(cl_mem), &clForce );
    error |= clSetKernelArg( clKernelBrownian, 4,
        sizeof(cl_ulong), &numParticles );

    // get good global item size
    clKernelBrownianGlobalSize = getKernelPaddedGlobalSize(
        sim_deviceId, clKernelBrownian, numParticles );

    // for debugging integers, until we initialize set false
    profileIntegers = false;
}

forceRandomGPU::~forceRandomGPU( )
{
    // release allocated memory
    // release kernels first
    clReleaseKernel( clKernelBrownian );
    if (profileIntegers) {
        // release allocated kernel for profiling integers
        clReleaseKernel( clKernelIntegers );
        // release allocated memory for profiling integers
        clReleaseMemObject( clRandomIntegers );
    }
    // release memory
    clReleaseMemObject( clRandomKey );
    clReleaseMemObject( clRandomCounter );
}

// calculating the force
void forceRandomGPU::calculateForce( )
{
    // run the kernel to get the forces
    checkErr(
        clEnqueueNDRangeKernel( // run the kernel
            queue, // we are adding it to our command queue
            clKernelBrownian, // the kernel we are running
            1, // 1d work space
            NULL, // no offsets
            &clKernelBrownianGlobalSize, // the global size
            NULL, // local items chosen by implementation
            0, NULL, NULL // we aren't waiting for any events
            ),
        "clEnqueueNDRangeKernel" );

    // the host and device arrays are no longer in synchronization
    updated = false;
}

// initialize details for debugging integer information
void forceRandomGPU::initializeProfileIntegers( )
{
    // we are profiling integers
    profileIntegers = true;

    // integer for obtaining errors
    cl_int error = 0;

    // create memory buffer for it
    clRandomIntegers = clCreateBuffer( context,
        CL_MEM_WRITE_ONLY, // make this buffer write only
        numParticles * sizeof(cl_int4),
        NULL, &error );
    checkErr( error, "clCreateBuffer (clRandomIntegers)" );

    // zero it out
    cl_int4 * zeroInts = new cl_int4[numParticles];
    for ( cl_ulong i = 0; i < numParticles; ++i ) {
        for ( int j = 0; j < 4; ++j ) {
            zeroInts[i].s[j] = 0;
        }
    }
    checkErr(
        clEnqueueWriteBuffer( queue, clRandomIntegers, CL_TRUE, 0,
            numParticles * sizeof( cl_int4 ),
            zeroInts, 0, NULL, NULL ),
        "clEnqueueWriteBuffer (clRandomIntegers)" );

    // create kernel for getting the random integers
    clKernelIntegers = clCreateKernel( program, "getRandomIntegers", &error );
    checkErr( error, "clCreateKernel (clKernelIntegers)" );

    // pass arguments into kernel
    error = clSetKernelArg( clKernelIntegers, 0,
        sizeof(cl_mem), &clRandomKey );
    error |= clSetKernelArg( clKernelIntegers, 1,
        sizeof(cl_mem), &clRandomCounter );
    error |= clSetKernelArg( clKernelIntegers, 2,
        sizeof(cl_mem), &clRandomIntegers );
    error |= clSetKernelArg( clKernelIntegers, 3,
        sizeof(cl_ulong), &numParticles );
}

// calculate random integers instead of the forces
void forceRandomGPU::generateIntegers( )
{
    if ( !profileIntegers ) {
        initializeProfileIntegers( );
    }

    // run the kernel to get the integers
    checkErr(
            clEnqueueNDRangeKernel( // run the kernel
            queue, // we are adding it to our command queue
            clKernelIntegers, // the kernel we are running
            1, // 1d work space
            NULL, // no offsets
            &clKernelBrownianGlobalSize, // the global size
            NULL, // local items chosen by implementation
            0, NULL, NULL // we aren't waiting for any events
            ),
        "clEnqueueNDRangeKernel (clKernelIntegers)" );
}

// save generated integers to chosen file
bool forceRandomGPU::saveIntegers(std::string fileName, bool append)
{
    if ( !profileIntegers ) {
        initializeProfileIntegers( );
    }

    cl_int4 * numbers = new cl_int4[numParticles];
    // read out the current integers
    checkErr(
        clEnqueueReadBuffer( queue, clRandomIntegers, CL_TRUE, 0,
            numParticles * sizeof(cl_int4), numbers,
            0, NULL, NULL ),
        "clEnqueueReadBuffer (clRandomIntegers)" ) ;

    // output the random numbers
    std::ofstream outFile;
    if ( append )
        outFile.open( fileName, std::ios::out | std::ios::app );
    else
        outFile.open( fileName, std::ios::out );
    // check if open
    if ( !outFile.is_open( ) ) {
        std::cerr << "ERROR: Unable to open "
            << fileName << "." << std::endl;
	return false;
    }

    // save our data
    for ( cl_ulong i = 0; i < numParticles; ++i ) {
        for ( int j = 0; j < 4; ++j ) {
            outFile << numbers[i].s[j] << "\n";
        }
    }

    // flush and close
    outFile.flush( );
    outFile.close( );

    delete[] numbers;
    return true;
}
