// forceSprings.cl
// OpenCL file for forceSpringsCL
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II


#include <chromoTypes.cl>


typedef struct simSPRING {
    ulong m1, m2; // indices for particles spring attached to
    ulong ndx1, ndx2; // indices for force vector
    CS_REAL restLength;
    CS_REAL springConstant;
    CS_REAL springDrag;
} simSPRING;


__kernel void calculateSpringForce( // calculates the force exerted by each of the springs
    __global simSPRING * spring, // all the springs
    __global CS_REAL3 * position, // the position of all the particles
    __global CS_REAL3 * velocity, // the velocity of all the particles
    __global CS_REAL3 * fSpring, // the forces applied on the masses by the springs
    const ulong numParticles, // the total number of particles
    const ulong numSprings // the total number of springs
    ) 
{
    ulong ndx = get_global_id(0); // we can worry about work groups later

    if (ndx >= numSprings)
    {
        return; // consequence of setting global size to multiple of preferred local group size
    }

    // indices of the two attached masses
    ulong m1 = spring[ndx].m1;
    ulong m2 = spring[ndx].m2;
    ulong ndx1 = spring[ndx].ndx1;
    ulong ndx2 = spring[ndx].ndx2;

    // locations of the two masses
    CS_REAL3 v1 = position[m1];
    CS_REAL3 v2 = position[m2];

    // displacement vector between the two masses
    CS_REAL3 d = v2 - v1;

    // calculate the force on mass 1
    CS_REAL3 f = 
        spring[ndx].springConstant * ( d - spring[ndx].restLength * normalize( d ) )
        - spring[ndx].springDrag * (velocity[m1] - velocity[m2]);

    // put into our two forces
    fSpring[m1 + ndx1*numParticles] = f;
    fSpring[m2 + ndx2*numParticles] = -f;
}

__kernel void calculateSpringForceAndJacobian( // calculate the force and Jacobian from the springs
    __global simSPRING * spring,
    __global CS_REAL3 * position,
    __global CS_REAL3 * velocity,
    __global CS_REAL3 * fSpring,
    __global CS_REAL3 * springJPosX,
    __global CS_REAL3 * springJPosY,
    __global CS_REAL3 * springJPosZ, // for now we are neglecting spring damping (separate class)
    const ulong numParticles,
    const ulong numSprings
    )
{
    ulong ndx = get_global_id(0); // we can worry about work groups later

    if (ndx >= numSprings)
    {
        return; // consequence of setting global size to multiple of preferred local group size
    }

    // indices of the two attached masses
    ulong m1 = spring[ndx].m1;
    ulong m2 = spring[ndx].m2;
    ulong ndx1 = spring[ndx].ndx1;
    ulong ndx2 = spring[ndx].ndx2;
    CS_REAL L = spring[ndx].restLength;
    CS_REAL k = spring[ndx].springConstant;

    // locations of the two masses
    CS_REAL3 v1 = position[m1];
    CS_REAL3 v2 = position[m2];

    // displacement vector between the two masses
    CS_REAL3 d = v2 - v1;
    CS_REAL dLength = length(d);

    CS_REAL scale = 1 - L / dLength;

    CS_REAL3 outerX = d.x * d / (dLength*dLength);
    CS_REAL3 outerY = d.y * d / (dLength*dLength);
    CS_REAL3 outerZ = d.z * d / (dLength*dLength);

    CS_REAL3 f = k * scale * d 
        - spring[ndx].springDrag * (velocity[m1] - velocity[m2]);
    CS_REAL3 jx = k * ( outerX + scale * ( ((float3) (1,0,0)) - outerX ) );
    CS_REAL3 jy = k * ( outerY + scale * ( ((float3) (0,1,0)) - outerY ) );
    CS_REAL3 jz = k * ( outerZ + scale * ( ((float3) (0,0,1)) - outerZ ) );

    // put into our two forces
    fSpring[m1 + ndx1*numParticles] = f;
    fSpring[m2 + ndx2*numParticles] = -f;

    // put into our Jacobian
    springJPosX[m1+(ndx1+1)*numParticles] = jx;
    springJPosY[m1+(ndx1+1)*numParticles] = jy;
    springJPosZ[m1+(ndx1+1)*numParticles] = jz;
    springJPosX[m2+(ndx2+1)*numParticles] = jx;
    springJPosY[m2+(ndx2+1)*numParticles] = jy;
    springJPosZ[m2+(ndx2+1)*numParticles] = jz;
}

__kernel void calculateSpringDiagonal( // calculate the diagonal elements of the Jacobian
    __global CS_REAL3 * springJPosX,
    __global CS_REAL3 * springJPosY,
    __global CS_REAL3 * springJPosZ, // for now we are neglecting spring damping (separate class)
    const ulong numParticles,
    const ulong maxInteractions // since we have fixed structure, we can just sum everything up
    )
{
    ulong ndx = get_global_id(0);

    if (ndx >= numParticles)
    {
        return;
    }

    // otherwise
    CS_REAL3 negativeSumX = - springJPosX[ndx+numParticles];
    CS_REAL3 negativeSumY = - springJPosY[ndx+numParticles];
    CS_REAL3 negativeSumZ = - springJPosZ[ndx+numParticles];
    for (int i = 2; i < maxInteractions; ++i)
    {
        negativeSumX -= springJPosX[ndx+i*numParticles];
        negativeSumY -= springJPosY[ndx+i*numParticles];
        negativeSumZ -= springJPosZ[ndx+i*numParticles];
    }
    springJPosX[ndx] = negativeSumX;
    springJPosY[ndx] = negativeSumY;
    springJPosZ[ndx] = negativeSumZ;
}