// genericForce.cpp
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include "genericForce.h"
#include "clErrorToString.h"
#include <math.h>

// constructor
genericForce::genericForce( 
    std::string sim_CLdir,
    cl_context sim_context, 
    cl_command_queue sim_queue, 
    cl_ulong sim_numParticles, 
    cl_ulong sim_totalMaxInteractions,
    bool inherited )
{
    // set CLdir to sim_CLdir
    CLdir = sim_CLdir;
    // set context to sim_context
    context = sim_context;
    // set queue to sim_queue
    queue = sim_queue;
    // set numParticles to sim_numParticles
    numParticles = sim_numParticles;
    // set totalMaxInteractions to sim_totalMaxInteractions (unless that's greater than numParticles)
    totalMaxInteractions = sim_totalMaxInteractions;
    if (numParticles < totalMaxInteractions) {
        totalMaxInteractions = numParticles;
    }
    // set clForceLoaded, clJacobianLoaded, and programLoaded to their default value of false
    clForceLoaded = false;
    clJacobianLoaded = false;
    programLoaded = false;
    // we have no interactions in Jacobian (by default)
    jacobianMaxInteractions = 0;

    if (!inherited) {
        // we only have one column
        columns = 1;
        // set the forces to zero
        initForcesZero( );
    }
}

// set up the Jacobian and set to zero, given the number of maximum interactions and numParticles
void genericForce::initJacobianZero( )
{
    if (jacobianMaxInteractions == 0) {
        return;
    }
    // totalMaxInteractions is total number of particles that a single particle is allowed to interact with at a time,
    // at least implicitly in the Jacobian.
    // jacobianMaxInteractions is the maximum number of interactions that might come out of this specific force. If
    // it is higher than totalMaxInteractions, then we might interact with too many particles. So we cannot let this happen.
    if (jacobianMaxInteractions > totalMaxInteractions) {
        // not okay.
        std::cerr << getName() << " will be treated explicitly" << std::endl;
        jacobianMaxInteractions = 0;
        return;
    }
    CS_REAL3 * zeroJ = new CS_REAL3[ numParticles * jacobianMaxInteractions ];
    cl_ulong * zeroNdx = new cl_ulong[ numParticles * jacobianMaxInteractions ];
    for (int i = 0; i < numParticles * jacobianMaxInteractions; ++i) {
        zeroNdx[i] = 0; // not fair to particle zero, but doesn't matter since no interactions when zero
        for (int j = 0; j < 3; ++j) {
            zeroJ[i].s[j] = 0;
        }
    }
    cl_ulong * zeroInteractionsCt = new cl_ulong[ numParticles ];
    for (int i = 0; i < numParticles; ++i) {
        zeroInteractionsCt[i] = 0;
    }
    // get errors from OpenCL API
    cl_int error;

    // set up buffers for Jacobian
    clJacobian.j_PosX = clCreateBuffer( 
        context, CL_MEM_READ_WRITE, 
        numParticles * jacobianMaxInteractions * sizeof(CS_REAL3), 
        NULL, &error );
    checkErr( error, "clCreateBuffer (j_PosX)" );
    clJacobian.j_PosY = clCreateBuffer( 
        context, CL_MEM_READ_WRITE, 
        numParticles * jacobianMaxInteractions * sizeof(CS_REAL3), 
        NULL, &error );
    checkErr( error, "clCreateBuffer (j_PosY)" );
    clJacobian.j_PosZ = clCreateBuffer( 
        context, CL_MEM_READ_WRITE, 
        numParticles * jacobianMaxInteractions * sizeof(CS_REAL3), 
        NULL, &error );
    checkErr( error, "clCreateBuffer (j_PosZ)" );
    clJacobian.j_VelX = clCreateBuffer( 
        context, CL_MEM_READ_WRITE, 
        numParticles * jacobianMaxInteractions * sizeof(CS_REAL3), 
        NULL, &error );
    checkErr( error, "clCreateBuffer (j_VelX)" );
    clJacobian.j_VelY = clCreateBuffer( 
        context, CL_MEM_READ_WRITE, 
        numParticles * jacobianMaxInteractions * sizeof(CS_REAL3), 
        NULL, &error );
    checkErr( error, "clCreateBuffer (j_VelY)" );
    clJacobian.j_VelZ = clCreateBuffer( 
        context, CL_MEM_READ_WRITE, 
        numParticles * jacobianMaxInteractions * sizeof(CS_REAL3), 
        NULL, &error );
    checkErr( error, "clCreateBuffer (j_VelZ)" );
    // set up buffer for the interaction indices
    clJacobian.interactionParticleIndex = clCreateBuffer( 
        context, CL_MEM_READ_WRITE, 
        numParticles * jacobianMaxInteractions * sizeof(cl_ulong), 
        NULL, &error );
    checkErr( error, "clCreateBuffer (interactionParticleIndex)" );
    // set up buffer for the number of interactions for given particle
    clJacobian.numInteractions = clCreateBuffer(
        context, CL_MEM_READ_WRITE,
        numParticles * sizeof(cl_ulong),
        NULL, &error );
    checkErr( error, "clCreateBuffer (numInteractions)" );

    // zero out the buffers
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clJacobian.j_PosX, // write to j_PosX
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), // size of the memory we are sending
            zeroJ, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (j_PosX)"
        );
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clJacobian.j_PosY, // write to j_PosY
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), // size of the memory we are sending
            zeroJ, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (j_PosY)"
        );
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clJacobian.j_PosZ, // write to j_PosZ
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), // size of the memory we are sending
            zeroJ, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (j_PosZ)"
        );
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clJacobian.j_VelX, // write to j_VelX
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), // size of the memory we are sending
            zeroJ, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (j_VelX)"
        );
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clJacobian.j_VelY, // write to j_VelY
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), // size of the memory we are sending
            zeroJ, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (j_VelY)"
        );
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clJacobian.j_VelZ, // write to j_VelZ
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * jacobianMaxInteractions * sizeof( CS_REAL3 ), // size of the memory we are sending
            zeroJ, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (j_VelZ)"
        );
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clJacobian.interactionParticleIndex, // write to interactionParticleIndex
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * jacobianMaxInteractions * sizeof( cl_ulong ), // size of the memory we are sending
            zeroNdx, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (interactionParticleIndex)"
        );
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clJacobian.numInteractions, // write to numInteractions
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            numParticles * sizeof( cl_ulong ), // size of the memory we are sending
            zeroInteractionsCt, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer (numInteractions)"
        );

    // let the genericForce know that we've allocated this space
    clJacobianLoaded = true;
}

// getter for the Jacobian
forceJacobian genericForce::getCLJacobian( )
{
    return clJacobian;
}

// the max number of nonzero elements for this force
cl_ulong genericForce::getJacobianMaxInteractions( )
{
    return jacobianMaxInteractions;
}

// initialize the forces to zero based on the columns and the numParticles
void genericForce::initForcesZero( )
{
    // initialize forces on host
    forces = new CS_REAL3[ numParticles * columns ];
    // set forces to zero
    for (int i = 0; i < numParticles * columns; ++i) {
        forces[i].s[0] = 0;
        forces[i].s[1] = 0;
        forces[i].s[2] = 0;
    }

    cl_int error = 0;
    // create buffer for clForce:
    clForce = clCreateBuffer( context, CL_MEM_READ_WRITE,
        columns * numParticles * sizeof( CS_REAL3 ), NULL, &error );
    clForceLoaded = true; // we will eventually need to unload the force

    updateDeviceForces( );
}

void genericForce::updateDeviceForces( )
{
    cl_int error = 0;
    // set to clForce the values of force
    checkErr(
        clEnqueueWriteBuffer( // write to buffer
            queue, // use the command queue
            clForce, // write to clForce
            CL_TRUE, // this is a blocking operation
            0, // no offset for the buffer
            columns * numParticles * sizeof( CS_REAL3 ), // size of the memory we are sending
            forces, // the array we are sending over
            0, NULL, NULL // we aren't using a wait list
            ),
        "clEnqueueWriteBuffer"
        );

    // since the device and host values for the forces should be the same, updated is true
    updated = true;
}

// function for updating clForce
void genericForce::calculateForce( )
{
    // Everything is zero, why change anything?
}

void genericForce::calculateForceAndJacobian( )
{
    // even though everything is zero, we must call calculateForce as default
    // behavior if this is called. This means that if we have a derived class that
    // overrided calculateForce but not calculateForceAndJacobian, we at least
    // get a force out of it.
    calculateForce( );
}

// function for the name of the force
const char * genericForce::getName( )
{
    return "Force Generic";
}

// destructor
genericForce::~genericForce( )
{
    delete[] forces;
    // these last two may have not actually been created.
    // However, they should just return a non CL_SUCCESS value,
    // and we can get along without any problems.
    if ( clForceLoaded ) {
        clReleaseMemObject( clForce );
    }
    if ( clJacobianLoaded ) {
        clReleaseMemObject( clJacobian.j_PosX );
        clReleaseMemObject( clJacobian.j_PosY );
        clReleaseMemObject( clJacobian.j_PosZ );
        clReleaseMemObject( clJacobian.j_VelX );
        clReleaseMemObject( clJacobian.j_VelY );
        clReleaseMemObject( clJacobian.j_VelZ );
        clReleaseMemObject( clJacobian.numInteractions );
        clReleaseMemObject( clJacobian.interactionParticleIndex );
    }
    if ( programLoaded ) {
        clReleaseProgram( program );
    }
}

// getter for the force pointer
cl_mem genericForce::getCLForce( )
{
    return clForce;
}

// getter for the number of columns
cl_ulong genericForce::getColumns( )
{
    return columns;
}

// retrieve the force from the host device
void genericForce::updateHostForces ( )
{
    checkErr(
        clEnqueueReadBuffer( queue, clForce, CL_TRUE, 0,
            numParticles * columns * sizeof( CS_REAL3 ), forces,
            0, NULL, NULL ),
        getName( ) );
    // they've been updated
    updated = true;
}

// print out the details of the force
std::string genericForce::printForce( )
{
    std::ostringstream ret_stream;
    if (!updated) {
        updateHostForces( );
    }
    for (cl_ulong s = 0; s < columns; ++s) {
        if (s > 0) {
            ret_stream << std::endl;
        }
        ret_stream << getName( );
        if (columns > 1) {
            ret_stream << s;
        }
        ret_stream << "\n";
        for (cl_ulong i = 0; i < numParticles; ++i) {
            cl_ulong ndx = columns*i + s;
            ret_stream << forces[ndx].s[0] << " "
                << forces[ndx].s[1] << " "
                << forces[ndx].s[2] << "\n";
        }
    }

    return ret_stream.str();
}

// save state of force to a file
bool genericForce::saveState( std::string fileName, bool append )
{
    // open the file
    std::ofstream outFile;
    if (append)
        outFile.open( fileName, std::ios::out | std::ios::app );
    else
        outFile.open( fileName, std::ios::out );
    // check if open
    if (!outFile.is_open()) {
        std::cerr << "ERROR: Unable to open "
            << fileName << "." << std::endl;
	return false;
    }

    // print it and close it
    outFile << printForce( );
    outFile.flush();
    outFile.close();
    return true;
}

// returns true if any of the forces are NAN
bool genericForce::anyNAN( )
{
    if (!updated) {
        updateHostForces( );
    }
    for (cl_ulong s = 0; s < columns; ++s) {
        for (cl_ulong i = 0; i < numParticles; ++i) {
            cl_ulong ndx = columns*i + s;
            if (forces[ndx].s[0] != forces[ndx].s[0] 
                || forces[ndx].s[1] != forces[ndx].s[1] 
                || forces[ndx].s[2] != forces[ndx].s[2]) {
                return true;
            }
        }
    }

    return false;
}

CS_REAL genericForce::maximumMagnitude( )
{
    CS_REAL retval = 0;
    if (!updated)
    {
        updateHostForces( );
    }
    for (cl_ulong i = 0; i < columns * numParticles; ++i)
    {
        CS_REAL vi = 0;
        for (int j = 0; j < 3; ++j)
            vi += forces[i].s[j] * forces[i].s[j];
        if (vi > retval)
            retval = vi;
    }

    return sqrt( retval );
}

void genericForce::checkErr ( cl_int err, const char * name )
{
    if (err != CL_SUCCESS)
    {
        std::cerr << "ERROR: " << name << " in " << getName( )
            << " (" << clErrorToString(err) << ")" << std::endl;
        exit( EXIT_FAILURE );
    }
}

bool genericForce::loadProgram ( cl_device_id deviceId, const char * file )
{
    // construct path to source file using CLdir and file
    std::ostringstream path;
    path << CLdir << "/" << file;
    // read in source file
    std::ifstream sourceFile( path.str() );
    if ( !sourceFile.is_open() ) {
        std::cerr << "ERROR: Could not open " << path.str() << " for kernel functions." << std::endl;
	return false;
    }
    std::string sourceCode(
        std::istreambuf_iterator<char>(sourceFile),
        (std::istreambuf_iterator<char>()));
    const char * src = sourceCode.c_str();
    size_t length = sourceCode.length();
    // create the program and build it
    cl_int error;
    // create the program
    program = clCreateProgramWithSource( context, 1, // XXX: not described in documentation, but example used magic number 1
        &src, &length, &error );
    checkErr( error, "clCreateProgramWithSource" );
    
    // set up the include directory to ensure consistent typedefs in kernel code (CS_REAL, CS_REAL3)
    std::ostringstream options;
    options << "-I " << CLdir;

    error = clBuildProgram( program, 1, // the number of devices
        &deviceId, options.str().c_str(), 
        NULL, NULL // stuff for notifications?
        );
    if (error != CL_SUCCESS) {
        // First call to know the proper size
        size_t log_size;
        clGetProgramBuildInfo(program, deviceId, CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
        char *build_log = (char* )malloc((log_size+1));

        // Second call to get the log
        clGetProgramBuildInfo(program, deviceId, CL_PROGRAM_BUILD_LOG, log_size, build_log, NULL);
        build_log[log_size] = '\0';
        std::cerr << "--- Build log ---\n ";
        std::cerr << build_log << "\n";
        free(build_log);
    }
    checkErr( error, "clBuildProgram" );
    // we have loaded the program, so it will eventually need to be unloaded
    programLoaded = true;
    return true;
}

// returns a padded global size using kernel preferred local work group size multiple
size_t genericForce::getKernelPaddedGlobalSize( 
    cl_device_id sim_deviceId, // the device id
    cl_kernel userKernel, // the kernel we are doing this for
    size_t minimumSize // the minimum value for the global size
    )
{
    // get preferred work group size multiple
    size_t preferredSize = getKernelMinimumPreferredLocalSize( sim_deviceId, userKernel );

    // now, set the global size to be the smallest size_t larger than minimumSize
    // and a multiple of preferredSize
    size_t result = (preferredSize - ( minimumSize % preferredSize )) + minimumSize;
    return result;
}

size_t genericForce::getKernelMinimumPreferredLocalSize(
    cl_device_id sim_deviceId, // the device id
    cl_kernel userKernel // the kernel we are doing this for
    )
{
    // get preferred work group size multiple
    size_t preferredSize = 0;
    checkErr( clGetKernelWorkGroupInfo(
            userKernel, sim_deviceId, 
            CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE,
            sizeof( preferredSize ), &preferredSize,
            NULL ),
        "clGetKernelWorkGroupInfo" );

    return preferredSize;
}
