// genericForce.h
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif
#include "chromoTypes.h"
#include <iostream>
#include <string>
#include <sstream>
#include <fstream>

#ifndef GENERICFORCE_H
#define GENERICFORCE_H

typedef struct forceJacobian
{
    // say that N is the number of particles, k the max number of interactions
    // ndx = i + N*b, i-th particle, interaction index b (0 <= i < N, 0 <= j < k) (buffers j_* are length N*k)
    cl_mem j_PosX, j_PosY, j_PosZ; // the 3 rows of the Jacobian wrt position, for each particle, for each interaction
    cl_mem j_VelX, j_VelY, j_VelZ; // the 3 rows of the Jacobian wrt velocity
    // ndx corresponds to an interaction with which particle? (buffer particleIndex is length N*k)
    cl_mem interactionParticleIndex; // the particle/column associated with the specific interaction (one particle by index used to access (row), the value for the column)
    // how many valid interactions does particle i have? (buffer numInteractions is length N)
    cl_mem numInteractions; // the number of interactions for each particle 
} forceJacobian;

// base class for the force calculations.
// Note that while we call it a force, we will also automatically multiply our results by
// the time step, making this more accurately called an impulse per time step
class genericForce
{
protected :
    // information about the force
    cl_mem clForce; // the pointer to device memory that will hold the forces
    cl_ulong columns; // the number of columns (for each of the particles)

    // information about the Jacobian
    forceJacobian clJacobian;
    cl_ulong jacobianMaxInteractions;

    // the force in host memory
    CS_REAL3 * forces;
    bool updated; // has it been updated since the last calculateForce?

    // OpenCL details
    cl_context context;
    cl_command_queue queue; // the command queue
    cl_program program; // a program that may or may not be created

    // what has been loaded?
    // these are set automatically to false, until initForcesZero, initJacobianZero and loadProgram set them
    // to true. This is used by the destructor to know if it needs to delete it.
    bool clForceLoaded;
    bool clJacobianLoaded;
    bool programLoaded;

    std::string CLdir; // the path of the directory with our OpenCL source code

    // details about the simulation
    cl_ulong numParticles; // the number of particles
    cl_ulong totalMaxInteractions;

    // retrieve the force from device and put to host
    void updateHostForces ( );

    // take the host from host and put on device
    void updateDeviceForces ( );

    // set up the forces and set them to zero, given the number of columns and numParticles
    void initForcesZero ( );

    // set up the Jacobian and set to zero, given the number of maximum interactions and numParticles
    void initJacobianZero ( );

    // check for error in calling some OpenCL function
    void checkErr ( cl_int err, const char * name );

    // load a given .cl file to program and build
    bool loadProgram ( cl_device_id deviceId, const char * file );

    // returns a padded global size using kernel preferred local work group size multiple
    size_t getKernelPaddedGlobalSize( 
        cl_device_id sim_deviceId, // the device id
        cl_kernel userKernel, // the kernel we are doing this for
        size_t minimumSize // the minimum value for the global size
        );

    size_t getKernelMinimumPreferredLocalSize(
        cl_device_id sim_deviceId, // the device id
        cl_kernel userKernel // the kernel we are doing this for
        );

public :
    // constructor
    genericForce ( 
        std::string sim_CLdir, // the path of directory with our openCL source code
        cl_context sim_context, // OpenCL context passed in from simulation
        cl_command_queue sim_queue, // OpenCL queue passed in from simulation
        cl_ulong sim_numParticles, // number of particles in simulation
        cl_ulong sim_totalMaxInteractions, // total maximum number of interactions (per particle) in Jacobian
        bool inherited // hacky way of skipping memory allocation when inherited
        );

    // destructor
    virtual ~genericForce ( );

    // virtual function for updating clForce
    virtual void calculateForce ( ); // updates clForce, sets updated to false

    // virtual function for updating clForce and clJacobian
    virtual void calculateForceAndJacobian ( ); // updates clForce and clJacobian, sets updated to false

    // what is the name of this force
    virtual const char * getName ( );

    // The remaining functions should all be the same regardless of force...

    // getters for the force pointer
    cl_mem getCLForce ( );
    // the number of columns in the force
    cl_ulong getColumns ( );

    // getter for the Jacobian
    forceJacobian getCLJacobian ( );
    // the max number of nonzero elements for this force
    cl_ulong getJacobianMaxInteractions ( );

    // print the current values of the force
    std::string printForce ( );

    bool saveState( std::string fileName, bool append );
    
    // returns true if any of the forces are NAN
    bool anyNAN ( );

    // returns the maximum magnitude of any of the forces
    CS_REAL maximumMagnitude ( );
};

#endif
