// solverImplicitEuler.cl
// By Joseph K Aicher, Summer 2014 with Professor Russell M Taylor II

#include <chromoTypes.cl> // CS_REAL, CS_REAL3

// kernel to use the summed forces, Jacobians, and current velocity
// in order to create buffers for the ViennaCL objects for
// system of equations for velocity update
__kernel void generateSystem (
    // input time step
    const CS_REAL timeStep, // the time step of our simulation (dt)
    // input buffers
    __global CS_REAL * inverseMass, // the (dt / mass) per particle
    __global CS_REAL3 * velocity, // the current velocity of the particles
    __global CS_REAL3 * force, // the current force on the particles
    __global CS_REAL3 * jPosX, // the Jacobian wrt position for the x coordinate row
    __global CS_REAL3 * jPosY, // " " for the y coordinate row
    __global CS_REAL3 * jPosZ, // " " for the z coordinate row
    __global CS_REAL3 * jVelX, // the Jacobian wrt velocity for the x coordinate row
    __global CS_REAL3 * jVelY, // " " for the y coordinate row
    __global CS_REAL3 * jVelZ, // " " for the z coordinate row
    __global ulong * jInteractionIndex, // the particle that this interaction corresponds to
    // output buffers
    __global CS_REAL * systemMatrixElement, // the nonzero values of the system matrix
    __global uint * systemMatrixColumn, // the column indices for the elements of system matrix
    __global CS_REAL * rhsVector, // the right hand side of the system of equations we are solving
    // parameters necessary to give size of the buffers
    const ulong maxInteractions, // 3 * maxInteractions is the number of nonzero columns per row of systemMatrix
    const ulong numParticles // 3 * numParticles is the dimension of our system of equations
    )
{
    ulong ndx = get_global_id(0);

    if ( ndx >= numParticles ) {
        return;
    }

    CS_REAL invMass = inverseMass[ndx];

    CS_REAL3 rhs = 0; // the three coordinates for the right hand side 

    ulong numColumns = 3 * maxInteractions;
    // indices of rows in viennaCL buffers
    ulong rowIndexX = 3 * ndx;
    ulong rowIndexY = rowIndexX + 1;
    ulong rowIndexZ = rowIndexX + 2;
    // locations of the start of x row, y row, and z rows in the matrix buffers
    ulong xRow = rowIndexX * numColumns; 
    ulong yRow = rowIndexY * numColumns; 
    ulong zRow = rowIndexZ * numColumns;
    // loop through the interactions (interaction*numParticles + ndx for the input buffers)
    // set values of systemMatrixElement, systemMatrixColumn. Add component of dt*df/dx*v to rhs
    for ( ulong interaction = 0; interaction < maxInteractions; ++interaction ) {
        // chromoSim index is interaction*numParticles + ndx (from the Jacobian buffers)
        ulong chromoSim_ndx = interaction*numParticles + ndx;
        // index of interacting particle is:
        ulong interactingIndex = jInteractionIndex[chromoSim_ndx];

        // first, get the values for systemMatrixColumn
        ulong colIndexX = 3 * interactingIndex;
        ulong colIndexY = colIndexX + 1;
        ulong colIndexZ = colIndexX + 2;

        // first get the Jacobian wrt x multiplied by time step. 
        // We take this intermediate step because we use it to also increment the rhs.
        CS_REAL3 updateX = timeStep * jPosX[chromoSim_ndx];
        CS_REAL3 updateY = timeStep * jPosY[chromoSim_ndx];
        CS_REAL3 updateZ = timeStep * jPosZ[chromoSim_ndx];

        // update rhs using this ( this is essentially a sparse block matrix multiplication )
        CS_REAL3 interactingVelocity = velocity[interactingIndex];
        rhs.x += dot( updateX, interactingVelocity );
        rhs.y += dot( updateY, interactingVelocity );
        rhs.z += dot( updateZ, interactingVelocity );

        // intermediate step towards finish making the update ( add the velocity Jacobian, multiply by inverse mass)
        updateX = invMass * ( jVelX[chromoSim_ndx] + updateX );
        updateY = invMass * ( jVelY[chromoSim_ndx] + updateY );
        updateZ = invMass * ( jVelZ[chromoSim_ndx] + updateZ );

        // finish the update by taking either 1 or 0 (if self interaction) and subtracting calculated update
        if ( ndx == interactingIndex ) {
            updateX = ((CS_REAL3) (1, 0, 0)) - updateX;
            updateY = ((CS_REAL3) (0, 1, 0)) - updateY;
            updateZ = ((CS_REAL3) (0, 0, 1)) - updateZ;
        } else {
            updateX = -updateX;
            updateY = -updateY;
            updateZ = -updateZ;
        }

        // make changes to buffers
        // xrow
        systemMatrixColumn[3*interaction + xRow] = colIndexX;
        systemMatrixColumn[3*interaction + xRow + 1] = colIndexY;
        systemMatrixColumn[3*interaction + xRow + 2] = colIndexZ;
        systemMatrixElement[3*interaction + xRow] = updateX.x;
        systemMatrixElement[3*interaction + xRow + 1] = updateX.y;
        systemMatrixElement[3*interaction + xRow + 2] = updateX.z;
        // yrow
        systemMatrixColumn[3*interaction + yRow] = colIndexX;
        systemMatrixColumn[3*interaction + yRow + 1] = colIndexY;
        systemMatrixColumn[3*interaction + yRow + 2] = colIndexZ;
        systemMatrixElement[3*interaction + yRow] = updateY.x;
        systemMatrixElement[3*interaction + yRow + 1] = updateY.y;
        systemMatrixElement[3*interaction + yRow + 2] = updateY.z;
        // zrow
        systemMatrixColumn[3*interaction + zRow] = colIndexX;
        systemMatrixColumn[3*interaction + zRow + 1] = colIndexY;
        systemMatrixColumn[3*interaction + zRow + 2] = colIndexZ;
        systemMatrixElement[3*interaction + zRow] = updateZ.x;
        systemMatrixElement[3*interaction + zRow + 1] = updateZ.y;
        systemMatrixElement[3*interaction + zRow + 2] = updateZ.z;
        // done so move on to next interaction
    }
    // we've finished updating the system matrix

    // right now rhs = dt * df/dx * v.
    // we want rhs = (dt * M^-1) * (f + dt*(df/dx)*v)
    // we just need to finish updating the rhs.
    // XXX Code inconsistent with the comments -- dt is not multiplied on the left.
    //     Changing to multiply by dt (and also to remove the dt already there)
    //     caused way-wrong answers.  So I suspect that the code is correct and the
    //     comments are broken.  Or else there is a counter-bug elsewhere.
    rhs = invMass * ( force[ndx] + rhs );
    // apply to rhs buffer
    rhsVector[rowIndexX] = rhs.x;
    rhsVector[rowIndexY] = rhs.y;
    rhsVector[rowIndexZ] = rhs.z;
}

// kernel to take the velocity update we solved for and apply it to
// our current position and velocity
__kernel void applyVelocityUpdate (
    // input time step
    const CS_REAL timeStep, // the time step of our simulation (dt)
    // input buffer from ViennaCL
    __global CS_REAL * velocityUpdateVector, // the velocity update we solved for
    // output buffer to chromoSim buffers
    __global CS_REAL3 * position, // the position buffer from chromoSim
    __global CS_REAL3 * velocity, // the velocity buffer from chromoSim
    // parameter we need for size of buffers
    const ulong numParticles // the number of particles in the simulation
    )
{
    ulong ndx = get_global_id(0);
    if ( ndx >= numParticles ) {
        return;
    }

    // get velocity update
    ulong viennaCLIndex = 3 * ndx;
    CS_REAL3 dv; // the change in velocity
    // get dv components from velocityUpdateVector
    dv.x = velocityUpdateVector[viennaCLIndex];
    dv.y = velocityUpdateVector[viennaCLIndex + 1];
    dv.z = velocityUpdateVector[viennaCLIndex + 2];

    // get updated velocity in private memory
    CS_REAL3 newv = velocity[ndx] + dv;
    
    // update position and velocity in global memory
    position[ndx] = position[ndx] + newv * timeStep;
    velocity[ndx] = newv;
}
