/*
	This just takes the average over different sets of values from
	within the radius of gyration file computed, reporting the averages
	from different times within the trace.
*/

#include <stdio.h>
#include <math.h>
#include <vector>

int main(int argc, const char *argv[])
{
    // Get the input and output files (for now, just standard in and out)
    FILE *infile = stdin;
    FILE *outfile = stdout;

    // Skip the header line from the file.
    char    line[4096];
    line[sizeof(line)-1] = '\0';    // Ensure null-termination
    if (fgets(line, sizeof(line)-1, infile) == NULL) {
	fprintf(stderr,"Could not read header line.\n");
        return(-1);
    }

    // Read in all of the values of radius of gyration,
    // end-to-end length, and contour length
    // from the file.
    std::vector<double> rogs;
    std::vector<double> e2es;
    std::vector<double> cons;
    while (true) {
        // Read and parse the next line
        if (fgets(line, sizeof(line)-1, infile) == NULL) {
            // Done reading lines, we got them all.
            break;
        }
        double time, rog,e2e,con;
        if (sscanf(line, "%lf, %lf, %lf, %lf", &time, &rog, &e2e, &con) != 4) {
            // No more lines for this line
            break;
        }
        rogs.push_back(rog);
        e2es.push_back(e2e);
        cons.push_back(con);
    }
    fclose(infile);

    // Compute the overall mean and standard error of the whole
    // data set and print them.
    size_t i;
    double rog,e2e,con;
    fprintf(outfile, "Period, Mean RoG, RoG stddev, RoG stderr, Mean EndToEnd, EndToEnd stddev, EndToEnd stderr, Mean Contour, Contour stddev, Contour stderr\n");
    rog = e2e = con = 0;
    for (i = 0; i < rogs.size(); i++) {
	rog += rogs[i];
	e2e += e2es[i];
	con += cons[i];
    }
    size_t count = rogs.size();
    double rogmean = rog / count;
    double e2emean = e2e / count;
    double conmean = con / count;
    double rogdev = 0, e2edev = 0, condev = 0;
    double rogerr = 0, e2eerr = 0, conerr = 0;
    for (i = 0; i < rogs.size(); i++) {
	rogdev += (rogmean - rogs[i]) * (rogmean - rogs[i]);
	e2edev += (e2emean - e2es[i]) * (e2emean - e2es[i]);
	condev += (conmean - cons[i]) * (conmean - cons[i]);
    }
    rogdev = sqrt(rogdev / count);
    e2edev = sqrt(e2edev / count);
    condev = sqrt(condev / count);
    rogerr = rogdev / sqrt(count);
    e2eerr = e2edev / sqrt(count);
    conerr = condev / sqrt(count);
    fprintf(outfile, "Whole run, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf\n",
	rogmean, rogdev, rogerr, e2emean, e2edev, e2eerr,
	conmean, condev, conerr );

    // Compute the mean and standard error of the second half of the 
    // data set and print them.
    rog = e2e = con = 0;
    size_t mid = rogs.size() / 2;
    for (i = mid; i < rogs.size(); i++) {
	rog += rogs[i];
	e2e += e2es[i];
	con += cons[i];
    }
    count = rogs.size() - mid;
    rogmean = rog / count;
    e2emean = e2e / count;
    conmean = con / count;
    rogdev = 0, e2edev = 0, condev = 0;
    rogerr = 0, e2eerr = 0, conerr = 0;
    for (i = mid; i < rogs.size(); i++) {
	rogdev += (rogmean - rogs[i]) * (rogmean - rogs[i]);
	e2edev += (e2emean - e2es[i]) * (e2emean - e2es[i]);
	condev += (conmean - cons[i]) * (conmean - cons[i]);
    }
    rogdev = sqrt(rogdev / count);
    e2edev = sqrt(e2edev / count);
    condev = sqrt(condev / count);
    rogerr = rogdev / sqrt(count);
    e2eerr = e2edev / sqrt(count);
    conerr = condev / sqrt(count);
    fprintf(outfile, "Whole run, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf, %lf\n",
	rogmean, rogdev, rogerr, e2emean, e2edev, e2eerr,
	conmean, condev, conerr );

    fclose(outfile);
    return 0;
}
